/*  GuiTuner 0.05
 *  Copyright (c) 2001 Fabio Checconi <fchecconi@libero.it>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <algorithm>
#include "conf.h"
#include "fftcalc.h"
#include "utils.h"

namespace Tuner {
	const double IFact = 1e8;
	const double Fact3rd = 1.259921050;
	const double Fact5th = 1.498307077;

	const int Range = 4;
}

inline bool
Tuner::FftCalc::peakInRange(int a, int b, double *buf) throw()
{
	while(a <= b)
		if(isPeak(a++, buf))
			return true;

	return false;
}

inline bool
Tuner::FftCalc::isPeak(int i, double *buf) throw()
{
	if(buf[i] > buf[i - 1] && buf[i] > buf[i + 1])
		return true;

	return false;
}

inline void
Tuner::FftCalc::fillPeakBuf(FftPeak &max) throw()
{
	resetPeaks();

	for(int i = 1; i < outSize / 2; ++i) {
		if(isPeak(i, fftOut)) {
			addPeak(i, fftOut[i]);

			if(fftOut[i] > max.value) {
				max.index = i;
				max.value = fftOut[i];
			}
		}
	}
}

inline void
Tuner::FftCalc::resetPeaks() throw()
{
	for(int i = 0; i < peakBuf.size(); ++i)
		peakBuf[i].reset();
}

inline void
Tuner::FftCalc::addPeak(int i, double v) throw()
{
	CmpVal cmp;

	if(v > peakBuf[0].value) {
		peakBuf[0].value = v, peakBuf[0].index = i;
		std::sort(peakBuf.begin(), peakBuf.end(), cmp);
	}
}

void
Tuner::FftCalc::doFft(unsigned char *ptr) throw()
{
	if(!iGate && !cGate) {
		if(pInt)
			pInt->_calc(ptr, fftOut);
		else
			pChar->_calc(ptr, fftOut);
	} else {
		ByteOrder ord;
		switch(fmt) {
			case S16BE:
				ord = BIGENDIAN;
				break;
			case S16LE:
				ord = LITTLEENDIAN;
				break;
			default:
				;
		}
		if(pInt) {
			char2int(ptr, iBuf, points, ord);
			iGate->apply(iBuf, points);
			pInt->calc(iBuf, fftOut);
		} else {
			cGate->apply(ptr, points);
			pChar->_calc(ptr, fftOut);
		}
	}
}

double
Tuner::FftCalc::countMethod() throw()
{
	int index = 0, k, t;
	FftPeak max;
	fillPeakBuf(max);

	CmpIndex cmp;
	std::sort(peakBuf.begin(), peakBuf.end(), cmp);

	for(int i = 0; i < PeakNumber; ++i) {
		if(peakBuf[i].index == 0 || peakBuf[i].value < (max.value / IFact))
			continue;

		if(peakBuf[i].index == max.index )
			peakBuf[i].isMaxHarm = true;

		for(k = 2 * peakBuf[i].index; k < outSize / 2; k *= 2) {
			for(t = i + 1; t < PeakNumber; ++t) {
				if(peakBuf[t].index >= k - Range &&
					peakBuf[t].index <= k + Range) {
					if(peakBuf[t].index == max.index)
						peakBuf[i].isMaxHarm = true;
					peakBuf[i].harmCount++;
				}
			}
		}

		double kf = Fact3rd * peakBuf[i].index;
		k = static_cast<int>(kf);
		while(k < outSize / 2) {
			for(t = i + 1; t < PeakNumber; ++t) {
				if(peakBuf[t].index >= k - Range &&
					peakBuf[t].index <= k + Range ) {
					if(peakBuf[t].index == max.index)
						peakBuf[i].isMaxHarm = true;
					peakBuf[i].harmCount++;
				}
			}
			kf *= 2.0;
			k = static_cast<int>(kf);
		}

		kf = Fact5th * peakBuf[i].index;
		k = static_cast<int>(kf);
		while(k < outSize / 2) {
			for(t = i + 1; t < PeakNumber; ++t) {
				if(peakBuf[t].index >= k - Range &&
					peakBuf[t].index <= k + Range) {
					if(peakBuf[t].index == max.index)
						peakBuf[i].isMaxHarm = true;
					peakBuf[i].harmCount++;
				}
			}
			kf *= 2.0;
			k = static_cast<int>(kf);
		}
	}
	k = 0;
	for(int i = 0; i < PeakNumber; ++i) {
		if(peakBuf[i].harmCount >= k && peakBuf[i].isMaxHarm) {
			index = peakBuf[i].index;
			k = peakBuf[i].harmCount;
		}
	}
	if( index == 0 )
		index = max.index;

	return static_cast<double>(index) * static_cast<double>(rate) /
		static_cast<double>(points);
}

double
Tuner::FftCalc::maxMethod() throw()
{
	double max = 0;
	int ind = 0;

	for(int i = 1; i < outSize - 1; ++i)
		if(fftOut[i] > max)
			max = fftOut[i], ind = i;

	return static_cast<double>(ind) * static_cast<double>(rate) /
		static_cast<double>(points);
}

Tuner::FftCalc::FftCalc(int p, SamplingRate s, Format f, Window::WindowType t,
	FftCalcMode m, bool g, int gp, int gh, int gr, int ga) throw(std::bad_alloc):
	mode(m), points(p), wt(t), useGate(g), percent(gp), hold(gh), rel(gr), att(ga),
	fftOut(0), pInt(0), pChar(0), iGate(0), cGate(0), iBuf(0), peakBuf(PeakNumber)
{
	setAudioParams(s, f, true);

	switch(mode) {
		case MAX_SEARCH:
			modeFun = &FftCalc::maxMethod;
			break;
		case PEAK_SEARCH:
			modeFun = &FftCalc::countMethod;
	}
}

Tuner::FftCalc::~FftCalc()
{
	delete []fftOut;

	if(iGate)
		delete iGate;
	if(cGate)
		delete cGate;
	if(pInt)
		delete pInt;
	if(pChar)
		delete pChar;
	if(iBuf)
		delete []iBuf;
}

void
Tuner::FftCalc::setAudioParams(SamplingRate s, Format f, bool first)
{
	const int intCenter = 0, charCenter = 128, intMax = 32768, charMax = 127;

	if(fmt != f || first) {
		if(pInt)
			delete pInt;
		if(pChar)
			delete pChar;

		switch(f) {
			case S16BE:
				pInt = new Fft<int>(points, BIGENDIAN, wt);
				outSize = pInt->outputSize();
				break;
			case S16LE:
				pInt = new Fft<int>(points, LITTLEENDIAN, wt);
				outSize = pInt->outputSize();
				break;
			default:
				pChar = new Fft<unsigned char>(points, ByteOrder(0), wt);
				outSize = pChar->outputSize();
		}

		if(fftOut)
			delete fftOut;
		fftOut = new double[outSize];
	}

	if(fmt != f || rate != s || first) {
		rate = s;
		fmt = f;

		if(iGate)
			delete iGate;
		if(cGate)
			delete cGate;

		if(useGate) {
			switch(fmt) {
				case S16BE:
				case S16LE:
					iGate = new Gate<int>(s, intCenter, intMax *
						percent / 100, hold, rel, att);
					break;
				default:
					cGate = new Gate<unsigned char>(s, charCenter,
						charMax * percent / 100, hold, rel,
						att);
			} 
		}
	}

	if(iBuf)
		delete []iBuf;
	iBuf = (iGate && pInt) ? new int[points]: 0;
}

double
Tuner::FftCalc::operator()(unsigned char *ptr) throw()
{
	doFft(ptr);
	// Unstable release only...
	double r = (this->*modeFun)();
	return r > 40? r: 0;
}

