/*  GuiTuner 0.05
 *  Copyright (c) 2001 Fabio Checconi <fchecconi@libero.it>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "objfactory.h"
#include "conf.h"
#include "genconf.h"
#include "alsa.h"
#include "esdout.h"
#include "oss.h"
#include "fftcalc.h"
#include "mainthread.h"
#include "utils.h"

Tuner::ObjFactory::ObjFactory(Config::GenConf &c) throw():
	config(c), audio(0), audioPlay(0), calc(0), buffer(0)
{
}

Tuner::ObjFactory::~ObjFactory()
{
	if(audio)
		delete audio;
	if(audioPlay)
		delete audioPlay;
	if(calc)
		delete calc;
	if(buffer)
		delete buffer;
}

bool
Tuner::ObjFactory::setup(const AudioInfo *ai) throw(std::bad_alloc)
{
	calc->setAudioParams(ai->rate, ai->fmt);
	CalcInfo fft = config.getCalcInfo();
	buffer->resize(fft.points * (ai->fmt == U8 ? 1: 2));

	return true;
}

bool
Tuner::ObjFactory::update() throw(Error::AudioError, Error::CalcError, std::bad_alloc)
{
	createAudio();
	createCalc();
	createBuffer();

	return true;
}

Tuner::Audio *&
Tuner::ObjFactory::createAudio() throw(Error::AudioError, std::bad_alloc)
{
	if(audio)
		delete audio;

	config.getAudioInfo().mode = RECORD;
	config.getAudioInfo().fragNumber = 512;
	config.getAudioInfo().fragSize = 256;
	config.getAudioInfo().chn = MONO;

	switch(config.getAudioDriver()) {
#ifdef HAVE_LIBASOUND
		case ALSA_DRIVER:
			return audio = new Alsa();
#endif
		case OSS_DRIVER:
			return audio = new Oss();
		case ESD_DRIVER:
			return audio = new Esd();
	}
}

Tuner::Audio *&
Tuner::ObjFactory::createAudioPlay() throw(Error::AudioError, std::bad_alloc)
{
	if(audioPlay)
		delete audioPlay;

	config.getAudioInfoPlay().mode = PLAYBACK;
	config.getAudioInfoPlay().fragNumber = 512;
	config.getAudioInfoPlay().fragSize = 256;
	config.getAudioInfoPlay().chn = MONO;

	switch(config.getAudioDriverPlay()) {
#ifdef HAVE_LIBASOUND
		case ALSA_DRIVER:
			return audioPlay = new Alsa();
#endif
		case Tuner::OSS_DRIVER:
			return audioPlay = new Oss();
		case Tuner::ESD_DRIVER:
			return audioPlay = new Esd();
	}
}

Tuner::Calc *&
Tuner::ObjFactory::createCalc(const AudioInfo *ai) throw(Error::CalcError, std::bad_alloc)
{
	if(calc)
		delete calc;

	switch(config.getDetectionMode()) {
		case FFTCALC: {
			CalcInfo fft = config.getCalcInfo();
			AudioInfo audio = ai ? *ai: config.getAudioInfo();
			return calc = new FftCalc(fft.points, audio.rate, audio.fmt,
				fft.wtype,	fft.mode, fft.useGate, fft.percent, fft.hold,
				fft.rel, fft.att);
		}
	}
}

Tuner::Buffer<unsigned char> *&
Tuner::ObjFactory::createBuffer(const AudioInfo *ai) throw(std::bad_alloc)
{
	if(buffer)
		delete buffer;

	CalcInfo fft = config.getCalcInfo();
	AudioInfo audio = ai ? *ai: config.getAudioInfo();
	return buffer = new Buffer<unsigned char>
		(fft.points * (audio.fmt == U8 ? 1: 2));
}

